#include "max-in-sub-array.h"

#include<iostream>
#include<iomanip>
#include<cstdlib>
#include<ctime>

using namespace std;

template<typename T>
void make_random_array(T* ptr, size_t size) {
  for (size_t i = 0; i < size; i++)
    ptr[i] = static_cast<T>(rand());
}

template<typename T>
T make_random_integer(T a, T b)
{
  T result(a + rand() % (b - a + 1));
  return result;
}

double time()
{
  timespec t;
  clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &t);
  return t.tv_sec + 1e-9 * t.tv_nsec;
}

enum { RANDOM, MONOTONIC };

template<typename T, int options>
double benchmark_one_path(int buffergenerator, size_t buffer_size, size_t update_size, size_t draw_size, size_t draws_between_updates)
{
  tbuffer<options> b;

  T *array = new T[buffer_size];
  if (buffergenerator == RANDOM)
    make_random_array(array, buffer_size);
  else
    for (size_t i = 0; i < buffer_size; i++)
      array[i] = T(double(i) * T(-1) / buffer_size);

  T max = 0;
  if (buffergenerator == RANDOM) {
    for (size_t i = 0; i < buffer_size; i++)
      if (array[i] > max)
        max = array[i];
  }

  b.bufferData(array, buffer_size * sizeof(T));

  double minimum_run_time = 0.1;
  size_t repeat = 1;
  double run_time = 0.0;

  while (true) {

    double start_time = time();

    for (size_t i = 0; i < repeat; i++) {
      size_t update_offset = make_random_integer<size_t>(0, buffer_size - update_size);

      b.bufferSubData(update_offset * sizeof(T), array + update_offset, update_size * sizeof(T));

      for (size_t j = 0; j < draws_between_updates; j++) {
        size_t draw_offset = make_random_integer<size_t>(0, buffer_size - draw_size);
        T max_allowed =
           buffergenerator == RANDOM
                 ? max
                 : T(double(draw_offset + draw_size) * T(-1) / buffer_size);
        if (!b.validate(max_allowed, draw_offset, draw_size))
          abort();
      }
    }

    double end_time = time();

    run_time = end_time - start_time;

    if (run_time >= minimum_run_time ||
        repeat >= 100000000)
    {
      break;
    }
    else
    {
      repeat *= 4;
    }
  }


  delete[] array;

  return run_time / repeat  + (1 + draws_between_updates) * 1e-6;
}

void benchmark(int buffergenerator, size_t buffer_size, size_t update_size, size_t draw_size, size_t draws_between_updates, double* time, double* speedup)
{
  double smart = benchmark_one_path<unsigned short, SMART>(buffergenerator, buffer_size, update_size, draw_size, draws_between_updates);
  double naive = benchmark_one_path<unsigned short, NAIVE>(buffergenerator, buffer_size, update_size, draw_size, draws_between_updates);
  *time = smart;
  *speedup = naive/smart;
}

int main()
{
  cout.precision(3);
  cout << setw(14) << "buffer update" << setw(14) << "number of" << setw(14) << "draw call" << setw(14) << "speedup" << setw(14) << "speedup"<< endl;
  cout << setw(14) << "size" << setw(14) << "draw calls" << setw(14) << "size" << setw(14) << "(loose)" << setw(14) << "(tight)" << endl;

  size_t buffer_size = 1048576;
  for (size_t draws_between_updates = 1;
          draws_between_updates <= 65536;
          draws_between_updates *= 16)
  {
    for (size_t draw_size = 4; draw_size <= std::min(buffer_size, size_t(65536)); draw_size *= 4)
    {
      for (size_t update_size = 4;
           update_size <= std::min(buffer_size, draw_size * draws_between_updates);
           update_size *= 4)
      {
        double time, speedup_random, speedup_monotonic;
        benchmark(RANDOM,
                  buffer_size,
                  std::min(update_size, size_t(0.9 * buffer_size)),
                  std::min(draw_size, size_t(0.9 * buffer_size)),
                  draws_between_updates,
                  &time,
                  &speedup_random
                 );
        benchmark(MONOTONIC,
                  buffer_size,
                  std::min(update_size, size_t(0.9 * buffer_size)),
                  std::min(draw_size, size_t(0.9 * buffer_size)),
                  draws_between_updates,
                  &time,
                  &speedup_monotonic
                 );
        cout << setw(14) << update_size
             << setw(14) << draws_between_updates
             << setw(14) << draw_size
             << setw(14) << speedup_random
             << setw(14) << speedup_monotonic
             
             << endl;
      }
    }
  }
}
