#pragma once

#include<vector>
#include <cstdlib>
#include <cstring>
#include <cstdint>

#include <cassert>

using std::size_t;

enum { UNSIGNED_BYTE = 0, UNSIGNED_SHORT = 1 };
enum { SMART, NAIVE };
template<int options = SMART>
struct tbuffer;

const size_t leaf_group_level = 3;
const size_t leaf_group_size = 1 << leaf_group_level;
const size_t leaf_group_mask = leaf_group_size - 1;



template<typename T>
struct get_tree_for_type {};


  typedef tbuffer<SMART> buffer;

template<typename T>
struct tree
{
  std::vector<T> data;
  bool invalidated;
  size_t invalidated_first;
  size_t invalidated_last;
  buffer& buf;

  tree(buffer& b) : invalidated(false), buf(b) {}

  T global_max() const {
    assert(!invalidated);
    return data[1];
  }

  static size_t parent(size_t i) {
    return i >> 1;
  }

  static bool is_right(size_t i) {
    return i & 1;
  }

  static bool is_left(size_t i) {
    return !is_right(i);
  }

  static size_t left_child(size_t i) {
    return i << 1;
  }

  static size_t right_child(size_t i) {
    return sibling(left_child(i));
  }

  static size_t left_neighbour(size_t i, size_t distance = 1) {
    return i - distance;
  }

  static size_t right_neighbour(size_t i, size_t distance = 1) {
    return i + distance;
  }

  static size_t sibling(size_t i) {
    return i ^ 1;
  }

  bool validate_for_first_and_last(T max_allowed, size_t first, size_t last) {
    assert(!invalidated);

    size_t size = data.size() / 2;
    first += size;
    last += size;

    while (true) {
      if (last <= first) {
        if (last < first)
          return true;
        return data[first] <= max_allowed;
      }
      if (is_right(first)) {
        if (data[first] > max_allowed)
          return false;
        first = right_neighbour(first);
      }
      if (is_left(last)) {
        if (data[last] > max_allowed)
          return false;
        last = left_neighbour(last);
      }
      first = parent(first);
      last = parent(last);
    }
  }

  void resize_in_bytes(size_t buffer_size_in_bytes)
  {
    size_t size = buffer_size_in_bytes / sizeof(T);
    size_t size_rounded_up = ((size - 1) | leaf_group_mask) + 1;
    size_t number_of_useful_leaves = size_rounded_up >> leaf_group_level;
    size_t tree_top_level_size = 1;
    while (tree_top_level_size < number_of_useful_leaves)
      tree_top_level_size <<= 1;
    data.resize(2 * tree_top_level_size);
    size_t max_leaf_index = tree_top_level_size - 1;
    invalidated_first = std::min(invalidated_first, max_leaf_index);
    invalidated_last = std::min(invalidated_last, max_leaf_index);
    
      
  }

  void invalidate_bytes(size_t first_byte, size_t last_byte)
  {
    size_t first = (first_byte / sizeof(T)) >> leaf_group_level;
    size_t last = (last_byte / sizeof(T)) >> leaf_group_level;

    if (invalidated) {

      size_t old_update_size = invalidated_last - invalidated_first + 1;
      size_t new_update_size = last - first + 1;

      size_t combined_first = std::min(first, invalidated_first);
      size_t combined_last = std::max(last, invalidated_last);

      size_t combined_update_size = combined_last
                                  - combined_first
                                  + 1;

      if (combined_update_size > old_update_size + new_update_size)
      {
        update();
        invalidated = true;
        invalidated_first = first;
        invalidated_last = last;

      } else {
        invalidated_first = combined_first;
        invalidated_last = combined_last;
      }
    } else {
      invalidated = true;
      invalidated_first = first;
      invalidated_last = last;
    }
  }

  void update();

};

template<int options>
struct tbuffer {

  void *untyped_data;
  size_t size_in_bytes;
  tree<uint8_t> tree_uint8;
  tree<uint16_t> tree_uint16;

  template<typename T> tree<T>& tree_for_type()
  {
    return get_tree_for_type<T>::run(this);
  }

  tbuffer() : untyped_data(0), size_in_bytes(0), tree_uint8(*this), tree_uint16(*this) {}
  ~tbuffer() { free(untyped_data); }

  template<typename T>
  const T* typed_data() const { return static_cast<const T*>(untyped_data); }
  template<typename T>
  T* typed_data() { return static_cast<T*>(untyped_data); }

  template<typename T>
  const T& element(size_t i) const { return typed_data<T>()[i]; }
  template<typename T>
  T& element(size_t i) { return typed_data<T>()[i]; }

  template<typename T>
  bool validate(T max_allowed, size_t first, size_t count) {
    tree<T>& tree = tree_for_type<T>();
    const T* data = typed_data<T>();
    if (false) {
    } else {

      if (!size_in_bytes)
        return true;
      if (!count)
        return true;
      if (tree.invalidated)
        tree.update();
      if (tree.global_max() <= max_allowed)
      {

        return true;
      }
      size_t last = first + count - 1;
      while ((first & leaf_group_mask) && (first < last)) {
        if (data[first] > max_allowed)
          return false;
        first++;
      }
      while (((last & leaf_group_mask) != leaf_group_mask) && (first < last)) {
        if (data[last] > max_allowed)
          return false;
        last--;
      }
      if (first == last)
        return data[first] <= max_allowed;

      return tree.validate_for_first_and_last(
               max_allowed,
               first >> leaf_group_level,
               last >> leaf_group_level);
    }
    return true;
  }

  template<typename T>
  void bufferData(const std::vector<T>& d) {
    bufferData(d.data(), d.size() * sizeof(T));
  }

  void bufferData(const void* ptr, size_t buffer_size_in_bytes) {
    size_in_bytes = buffer_size_in_bytes;
    tree_uint8.resize_in_bytes(buffer_size_in_bytes);
    tree_uint16.resize_in_bytes(buffer_size_in_bytes);
    untyped_data = realloc(untyped_data, buffer_size_in_bytes);
    bufferSubData(0, ptr, buffer_size_in_bytes);
  }

  template<typename T>
  void bufferSubData(size_t pos, const std::vector<T>& d) {
    assert((pos + d.size()) * sizeof(T) <= size_in_bytes);
    bufferSubData(pos * sizeof(T), d.data(), d.size() * sizeof(T));
  }

  void bufferSubData(size_t pos, const void* ptr, size_t update_size_in_bytes) {
    if (!update_size_in_bytes) return;
    memcpy(static_cast<uint8_t*>(untyped_data) + pos, ptr, update_size_in_bytes);
    invalidate_trees(pos, pos + update_size_in_bytes - 1);
  }

  void invalidate_trees(size_t first_byte, size_t last_byte)
  {
    tree_uint8.invalidate_bytes(first_byte, last_byte);
    tree_uint16.invalidate_bytes(first_byte, last_byte);
  }

private:


};

template<>
struct tbuffer<NAIVE> {

  void *untyped_data;
  size_t size_in_bytes;

  tbuffer() : untyped_data(0), size_in_bytes(0) {}
  ~tbuffer() { free(untyped_data); }

  template<typename T>
  const T* typed_data() const { return static_cast<const T*>(untyped_data); }
  template<typename T>
  T* typed_data() { return static_cast<T*>(untyped_data); }

  template<typename T>
  const T& element(size_t i) const { return typed_data<T>()[i]; }
  template<typename T>
  T& element(size_t i) { return typed_data<T>()[i]; }

  template<typename T>
  bool validate(T max_allowed, size_t first, size_t count) {
    size_t last = first + count - 1;
    for (size_t i = first; i <= last; i++)
      if (element<T>(i) > max_allowed)
        return false;
    return true;
      
  }

  template<typename T>
  void bufferData(const std::vector<T>& d) {
    bufferData(d.data(), d.size() * sizeof(T));
  }

  void bufferData(const void* ptr, size_t buffer_size_in_bytes) {
    size_in_bytes = buffer_size_in_bytes;
    untyped_data = realloc(untyped_data, buffer_size_in_bytes);
    bufferSubData(0, ptr, buffer_size_in_bytes);
  }

  template<typename T>
  void bufferSubData(size_t pos, const std::vector<T>& d) {
    bufferSubData(pos, d.data(), d.size() * sizeof(T));
  }

  void bufferSubData(size_t pos, const void* ptr, size_t update_size_in_bytes) {
    if (!update_size_in_bytes) return;
    memcpy(static_cast<uint8_t*>(untyped_data) + pos, ptr, update_size_in_bytes);
  }
};

template<>
struct get_tree_for_type<uint8_t>
{
  static tree<uint8_t>& run(buffer *b) { return b->tree_uint8; }
};

template<>
struct get_tree_for_type<uint16_t>
{
  static tree<uint16_t>& run(buffer *b) { return b->tree_uint16; }
};

template<typename T>
void tree<T>::update()
{
    assert(invalidated);

    invalidated = false;

    if (data.size() == 0)
      return;

    size_t tree_first = data.size() / 2 + invalidated_first;
    size_t tree_last = data.size() / 2 + invalidated_last;

    {
      size_t tree_index = tree_first;
      size_t buffer_index = invalidated_first * leaf_group_size;
      while (tree_index <= tree_last) {
        T m = 0;
        for (size_t i = 0; i < leaf_group_size; i++) {
          m = std::max(m, buf.element<T>(buffer_index++));
        }
        data[tree_index] = m;
        tree_index++;
      }
    }

    size_t first = tree_first;
    size_t last = tree_last;

    while (true) {
      if (first == last) {
        size_t firstparent = parent(first);
        while (firstparent) {
          data[firstparent] = std::max(data[first], data[sibling(first)]);
          first = firstparent;
          firstparent = parent(first);
        }
        break;
      }
      first = parent(first);
      last = parent(last);
      size_t child = left_child(first);
      size_t parent = first;

      const int unroll_size = 8;
      while (right_neighbour(parent, unroll_size - 1) <= last)
      {
        asm("#begin");
        for (int unroll = 0; unroll < unroll_size; unroll++)
                    {
            T a = data[child];
            child = right_neighbour(child);
            T b = data[child];
            child = right_neighbour(child);
            data[parent] = std::max(a, b);
            parent = right_neighbour(parent);
          }
        asm("#end");
      }
      while (parent <= last)
      {
         {
            T a = data[child];
            child = right_neighbour(child);
            T b = data[child];
            child = right_neighbour(child);
            data[parent] = std::max(a, b);
            parent = right_neighbour(parent);
          }
      }
    }

  }
